package sun.nio.ch;

import java.util.Iterator;
import java.nio.channels.AsynchronousCloseException;
import java.util.HashMap;
import java.util.Map;
import sun.misc.Unsafe;

class PendingIoCache
{
    private static final Unsafe unsafe;
    private static final int addressSize;
    private static final int SIZEOF_OVERLAPPED;
    private boolean closed;
    private boolean closePending;
    private final Map<Long, PendingFuture> pendingIoMap;
    private long[] overlappedCache;
    private int overlappedCacheCount;
    
    private static int dependsArch(int n, int n2) {
        return (PendingIoCache.addressSize == 4) ? n : n2;
    }
    
    PendingIoCache() {
        this.pendingIoMap = new HashMap();
        this.overlappedCache = new long[4];
        this.overlappedCacheCount = 0;
    }
    
    long add(PendingFuture<?, ?> pendingFuture) {
        synchronized (this) {
            if (this.closed) {
                throw new AssertionError("Should not get here");
            }
            long allocateMemory;
            if (this.overlappedCacheCount > 0) {
                long[] overlappedCache = this.overlappedCache;
                int overlappedCacheCount = this.overlappedCacheCount - 1;
                this.overlappedCacheCount = overlappedCacheCount;
                allocateMemory = overlappedCache[overlappedCacheCount];
            }
            else {
                allocateMemory = PendingIoCache.unsafe.allocateMemory((long)PendingIoCache.SIZEOF_OVERLAPPED);
            }
            this.pendingIoMap.put(Long.valueOf(allocateMemory), pendingFuture);
            return allocateMemory;
        }
    }
    
     <V, A> PendingFuture<V, A> remove(long n) {
        synchronized (this) {
            PendingFuture<Object, Object> pendingFuture = (PendingFuture)this.pendingIoMap.remove(Long.valueOf(n));
            if (pendingFuture != null) {
                if (this.overlappedCacheCount < this.overlappedCache.length) {
                    this.overlappedCache[this.overlappedCacheCount++] = n;
                }
                else {
                    PendingIoCache.unsafe.freeMemory(n);
                }
                if (this.closePending) {
                    this.notifyAll();
                }
            }
            return pendingFuture;
        }
    }
    
    void close() {
        synchronized (this) {
            if (this.closed) {
                return;
            }
            if (!this.pendingIoMap.isEmpty()) {
                this.clearPendingIoMap();
            }
            while (this.overlappedCacheCount > 0) {
                Unsafe unsafe = PendingIoCache.unsafe;
                long[] overlappedCache = this.overlappedCache;
                int overlappedCacheCount = this.overlappedCacheCount - 1;
                this.overlappedCacheCount = overlappedCacheCount;
                unsafe.freeMemory(overlappedCache[overlappedCacheCount]);
            }
            this.closed = true;
        }
    }
    
    private void clearPendingIoMap() {
        assert !(!Thread.holdsLock(this));
        this.closePending = true;
        try {
            this.wait(50L);
        }
        catch (InterruptedException ex) {
            Thread.currentThread().interrupt();
        }
        this.closePending = false;
        if (this.pendingIoMap.isEmpty()) {
            return;
        }
        for (Long n : this.pendingIoMap.keySet()) {
            PendingFuture<Object, Object> pendingFuture = (PendingFuture)this.pendingIoMap.get(n);
            assert !pendingFuture.isDone();
            Iocp iocp = (Iocp)((Groupable)pendingFuture.channel()).group();
            iocp.makeStale(n);
            iocp.executeOnPooledThread(new Runnable() {
                
                @Override
                public void run() {
                    rh.failed(-1, new AsynchronousCloseException());
                }
            });
        }
        this.pendingIoMap.clear();
        return;
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        addressSize = PendingIoCache.unsafe.addressSize();
        SIZEOF_OVERLAPPED = dependsArch(20, 32);
    }
}
